<nav class="border-gray-200 bg-gray-50 dark:border-gray-700 dark:bg-gray-800">
    <div class="max-w-(--breakpoint-xl) mx-auto flex flex-wrap items-center justify-between p-4">
        <a href="/" class="flex items-center space-x-3 rtl:space-x-reverse">
            <img src="favicon.png" class="logo h-10" alt="Logo" />
            <span class="self-center whitespace-nowrap text-2xl font-semibold dark:text-[#d79921]">
                Qlue-ls
            </span>
        </a>
        <div>
            <a
                href="https://github.com/IoannisNezis/sparql-language-server"
                class="flex flex-row p-2 text-white hover:text-gray-400"
            >
                <img class="h-6 w-6" src="/Git-Icon-White.svg" alt="" />
                <span class="ms-4"> IoannisNezis/Qlue-ls </span>
            </a>
        </div>
    </div>
</nav>
